/* Boost interval/rounded_arith.hpp template implementation file
 *
 * Copyright 2002-2003 Herv Brnnimann, Guillaume Melquiond, Sylvain Pion
 *
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file LICENSE_1_0.txt or
 * copy at http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef BOOST_NUMERIC_INTERVAL_ROUNDED_ARITH_HPP
#define BOOST_NUMERIC_INTERVAL_ROUNDED_ARITH_HPP

#include <boost/numeric/interval/rounding.hpp>
#include <boost/numeric/interval/detail/bugs.hpp>
#include <boost/config/no_tr1/cmath.hpp>

namespace boost {
namespace numeric {
namespace interval_lib {

/*
 * Three classes of rounding: exact, std, opp
 * See documentation for details.
 */

template<class T, class Rounding>
struct rounded_arith_exact: Rounding {
  void init() { }
  template<class U> T conv_down(U const &v) { return v; }
  template<class U> T conv_up  (U const &v) { return v; }
  T add_down (const T& x, const T& y) { return x + y; }
  T add_up   (const T& x, const T& y) { return x + y; }
  T sub_down (const T& x, const T& y) { return x - y; }
  T sub_up   (const T& x, const T& y) { return x - y; }
  T mul_down (const T& x, const T& y) { return x * y; }
  T mul_up   (const T& x, const T& y) { return x * y; }
  T div_down (const T& x, const T& y) { return x / y; }
  T div_up   (const T& x, const T& y) { return x / y; }
  T median   (const T& x, const T& y) { return (x + y) / 2; }
  T sqrt_down(const T& x)
  { BOOST_NUMERIC_INTERVAL_using_math(sqrt); return sqrt(x); }
  T sqrt_up  (const T& x)
  { BOOST_NUMERIC_INTERVAL_using_math(sqrt); return sqrt(x); }
  T int_down (const T& x)
  { BOOST_NUMERIC_INTERVAL_using_math(floor); return floor(x); }
  T int_up   (const T& x)
  { BOOST_NUMERIC_INTERVAL_using_math(ceil); return ceil(x); }
};

template<class T, class Rounding>
struct rounded_arith_std: Rounding {
#define BOOST_FR(EXPR) this->force_rounding(EXPR)
  void init() { }
  template<class U> T conv_down(U const &v) { this->downward(); return BOOST_FR(v); }
  template<class U> T conv_up  (U const &v) { this->upward();   return BOOST_FR(v); }
  T add_down(const T& x, const T& y) { this->downward(); return BOOST_FR(BOOST_FR(x) + y); }
  T sub_down(const T& x, const T& y) { this->downward(); return BOOST_FR(BOOST_FR(x) - y); }
  T mul_down(const T& x, const T& y) { this->downward(); return BOOST_FR(BOOST_FR(x) * y); }
  T div_down(const T& x, const T& y) { this->downward(); return BOOST_FR(BOOST_FR(x) / y); }
  T add_up  (const T& x, const T& y) { this->upward();    return BOOST_FR(BOOST_FR(x) + y); }
  T sub_up  (const T& x, const T& y) { this->upward();    return BOOST_FR(BOOST_FR(x) - y); }
  T mul_up  (const T& x, const T& y) { this->upward();    return BOOST_FR(BOOST_FR(x) * y); }
  T div_up  (const T& x, const T& y) { this->upward();    return BOOST_FR(BOOST_FR(x) / y); }
  T median(const T& x, const T& y) { this->to_nearest(); return BOOST_FR((BOOST_FR(x) + y) / 2); }
  T sqrt_down(const T& x)
  { BOOST_NUMERIC_INTERVAL_using_math(sqrt); this->downward(); return BOOST_FR(sqrt(BOOST_FR(x))); }
  T sqrt_up  (const T& x)
  { BOOST_NUMERIC_INTERVAL_using_math(sqrt); this->upward();   return BOOST_FR(sqrt(BOOST_FR(x))); }
  T int_down(const T& x) { this->downward(); return this->to_int(x); }
  T int_up  (const T& x) { this->upward();   return this->to_int(x); }
#undef BOOST_FR
};
  
template<class T, class Rounding>
struct rounded_arith_opp: Rounding {
#define BOOST_FR(EXPR) this->force_rounding(EXPR)
  void init() { this->upward(); }
  template<class U> T conv_down(U const &v) { return -BOOST_FR(-v); }
  template<class U> T conv_up  (U const &v) { return BOOST_FR(v); }
  T add_down(const T& x, const T& y) { return -BOOST_FR(BOOST_FR(-x) - y); }
  T sub_down(const T& x, const T& y) { return -BOOST_FR(BOOST_FR(y) - x); }
  T mul_down(const T& x, const T& y) { return -BOOST_FR(BOOST_FR(x) * (-y)); }
  T div_down(const T& x, const T& y) { return -BOOST_FR(BOOST_FR(x) / (-y)); }
  T add_up  (const T& x, const T& y) { return BOOST_FR(BOOST_FR(x) + y); }
  T sub_up  (const T& x, const T& y) { return BOOST_FR(BOOST_FR(x) - y); }
  T mul_up  (const T& x, const T& y) { return BOOST_FR(BOOST_FR(x) * y); }
  T div_up  (const T& x, const T& y) { return BOOST_FR(BOOST_FR(x) / y); }
  T median  (const T& x, const T& y) { this->to_nearest(); T r = (BOOST_FR(x) + y) / 2; this->upward(); return BOOST_FR(r);}
  T sqrt_down(const T& x)
  { BOOST_NUMERIC_INTERVAL_using_math(sqrt); this->downward(); T r = sqrt(BOOST_FR(x)); this->upward(); return BOOST_FR(r);}
  T sqrt_up  (const T& x)
  { BOOST_NUMERIC_INTERVAL_using_math(sqrt); return BOOST_FR(sqrt(BOOST_FR(x))); }
  T int_down(const T& x) { return -this->to_int(-x); }
  T int_up  (const T& x) { return this->to_int(x); }
#undef BOOST_FR
};

} // namespace interval_lib
} // namespace numeric
} // namespace boost

#endif // BOOST_NUMERIC_INTERVAL_ROUNDED_ARITH_HPP
